#ifndef DUNE_CURVEDGRID_ANALYTIC_GRIDFUNCTION_HH
#define DUNE_CURVEDGRID_ANALYTIC_GRIDFUNCTION_HH

#include <optional>
#include <type_traits>
#include <utility>

#include <dune/common/typeutilities.hh>

#include "gridentityset.hh"

namespace Dune {

/// \brief LocalFunction associated to the \ref AnalyticGridFunction
/**
 * \tparam LocalContext  Context this localFunction can be bound to, e.g. a grid element
 * \tparam F             Type of a function that can be evaluated in global coordinates.
 **/
template <class LocalContext, class F>
class LocalAnalyticGridFunction;

/// \brief Generator for \ref LocalAnalyticGridFunction
/**
 * \param  ff            Function that can be evaluated at global coordinates of the \ref LocalContext
 * \tparam LocalContext  Context this localFunction can be bound to, e.g. a grid element
 **/
template <class LocalContext, class FF>
auto localAnalyticGridFunction (FF&& ff)
{
  using F = std::decay_t<FF>;
  return LocalAnalyticGridFunction<LocalContext, F>{std::forward<FF>(ff)};
}


template <class LC, class Functor>
class LocalAnalyticGridFunction
{
public:
  using LocalContext = LC;
  using Geometry = typename LocalContext::Geometry;

  using Domain = typename Geometry::GlobalCoordinate;
  using LocalDomain = typename Geometry::LocalCoordinate;
  using Range = std::result_of_t<Functor(Domain)>;
  using Signature = Range(LocalDomain);

public:
  /// \brief Constructor. Stores the functor f by value
  template <class FF,
    disableCopyMove<LocalAnalyticGridFunction, FF> = 0>
  LocalAnalyticGridFunction (FF&& f)
    : f_(std::forward<FF>(f))
  {}

  /// \brief bind the LocalFunction to the local context
  /**
   * Stores the localContext and its geometry in a cache variable
   **/
  void bind (const LocalContext& localContext)
  {
    localContext_.emplace(localContext);
    geometry_.emplace(localContext.geometry());
  }

  /// \brief unbind the localContext from the localFunction
  /**
   * Reset the geometry
   **/
  void unbind ()
  {
    geometry_.reset();
    localContext_.reset();
  }

  /// \brief evaluate the stored function in local coordinates
  /**
   * Transform the local coordinates to global coordinates first,
   * then evalute the stored functor.
   **/
  Range operator() (const LocalDomain& x) const
  {
    assert(!!geometry_);
    return f_(geometry_->global(x));
  }

  /// \brief return the bound localContext.
  const LocalContext& localContext () const
  {
    assert(!!localContext_);
    return *localContext_;
  }

  /// \brief obtain the functor
  const Functor& f () const
  {
    return f_;
  }

private:
  Functor f_;

  // some caches
  std::optional<LocalContext> localContext_;
  std::optional<Geometry> geometry_;
};

/// \brief Derivative of a \ref LocalAnalyticGridFunction
/**
 * Participates in overload resolution only if the functor `F` is differentiable
 **/
template <class LC, class F>
auto derivative (const LocalAnalyticGridFunction<LC,F>& t)
  -> decltype(localAnalyticGridFunction<LC>(derivative(t.f())))
{
  return localAnalyticGridFunction<LC>(derivative(t.f()));
}


/// \brief GridFunction associated to the mapping F
/**
 * \tparam Grid       The grid type with elements the corresponding LocalFunction can be bound to
 * \tparam F          Type of a function that can be evaluated in global coordinates.
 **/
template <class Grid, class F>
class AnalyticGridFunction;

/// \brief Generator for \ref AnalyticGridFunction
/**
 * \param  ff    Function that can be evaluated at global coordinates of the \ref Grid
 * \tparam Grid  The grid type with elements the corresponding LocalFunction can be bound to
 **/
template <class Grid, class FF>
auto analyticGridFunction (FF&& ff)
{
  using F = std::decay_t<FF>;
  return AnalyticGridFunction<Grid, F>{std::forward<FF>(ff)};
}

template <class FF, class Grid>
auto analyticGridFunction (FF&& ff, Grid const& /*grid*/)
{
  using F = std::decay_t<FF>;
  return AnalyticGridFunction<Grid, F>{std::forward<FF>(ff)};
}

template <class GridType, class Functor>
class AnalyticGridFunction
{
public:
  using Grid = GridType;
  using EntitySet = GridEntitySet<Grid,0>;

  using Domain = typename EntitySet::GlobalCoordinate;
  using Range = std::result_of_t<Functor(Domain)>;
  using Signature = Range(Domain);

public:
  /// \brief Constructor. Stores the functor f by value
  template <class FF,
    disableCopyMove<AnalyticGridFunction, FF> = 0>
  AnalyticGridFunction (FF&& f)
    : f_(std::forward<FF>(f))
  {}

  /// \brief evaluate the stored function in global coordinates
  Range operator() (const Domain& x) const
  {
    return f_(x);
  }

  /// \brief construct the \ref LocalAnalyticGridFunction
  using LocalFunction = LocalAnalyticGridFunction<typename EntitySet::Element, Functor>;
  friend LocalFunction localFunction (const AnalyticGridFunction& t)
  {
    return LocalFunction(t.f_);
  }

  /// \brief obtain the stored \ref GridEntitySet
  const EntitySet& entitySet () const
  {
    return entitySet_;
  }

  /// \brief obtain the functor
  const Functor& f () const
  {
    return f_;
  }

private:
  Functor f_;
  EntitySet entitySet_;
};

/// \brief Derivative of an \ref AnalyticGridFunction
/**
 * Participates in overload resolution only if the functor `F` is differentiable
 **/
template <class Grid, class F>
auto derivative (const AnalyticGridFunction<Grid,F>& t)
  -> decltype(analyticGridFunction<Grid>(derivative(t.f())))
{
  return analyticGridFunction<Grid>(derivative(t.f()));
}

} // end namespace Dune

#endif // DUNE_CURVEDGRID_ANALYTIC_GRIDFUNCTION_HH
