#ifndef DUNE_CURVEDGRID_GRIDVIEW_HH
#define DUNE_CURVEDGRID_GRIDVIEW_HH

#include <cassert>
#include <utility>

#include <dune/curvedgrid/datahandle.hh>
#include <dune/curvedgrid/declaration.hh>
#include <dune/curvedgrid/indexsets.hh>
#include <dune/curvedgrid/intersection.hh>
#include <dune/curvedgrid/intersectioniterator.hh>
#include <dune/curvedgrid/iterator.hh>
#include <dune/grid/common/gridview.hh>

namespace Dune {
namespace Curved {

// Internal Forward Declarations
// -----------------------------

template <class HGV, class GF, bool useInterpolation>
class GridView;


// GridViewTraits
// --------------

template <class HGV, class GF, bool useInterpolation>
class GridViewTraits
{
  friend class GridView<HGV, GF, useInterpolation>;

  using HostGridView = HGV;
  using GridFunction = GF;

  using HostGrid = typename HostGridView::Grid;
  using HostIntersection = typename HostGridView::Intersection;
  using HostIntersectionIterator = typename HostGridView::IntersectionIterator;

public:
  using GridViewImp = GridView<HostGridView, GridFunction, useInterpolation>;
  using Grid = Dune::CurvedGrid<GridFunction, useInterpolation>;
  using IndexSet = Curved::IndexSet<const Grid, typename HostGridView::IndexSet>;

  using Intersection = Dune::Intersection<const Grid, Curved::Intersection<const Grid, HostIntersection>>;

  using IntersectionIterator
    = Dune::IntersectionIterator<const Grid,
        Curved::IntersectionIterator<const Grid, HostIntersectionIterator>,
        Curved::Intersection<const Grid, HostIntersection>>;

  using CollectiveCommunication = typename HostGridView::CollectiveCommunication;

  template <int codim>
  struct Codim
  {
    using IteratorImp = Curved::Iterator<HostGridView, codim, All_Partition, const Grid>;
    using Iterator = Dune::EntityIterator<codim, const Grid, IteratorImp>;

    using Entity = typename Grid::Traits::template Codim<codim>::Entity;

    using Geometry = typename Grid::template Codim<codim>::Geometry;
    using LocalGeometry = typename Grid::template Codim<codim>::LocalGeometry;

    template <PartitionIteratorType pit>
    struct Partition
    {
      using IteratorImp = Curved::Iterator<HostGridView, codim, pit, const Grid>;
      using Iterator = Dune::EntityIterator<codim, const Grid, IteratorImp>;
    };
  };

  static const bool conforming = HostGridView::conforming;
};



// GridView
// --------

template <class HGV, class GF, bool useInterpolation>
class GridView
{
  using Self = GridView;

public:
  using Traits = GridViewTraits<HGV, GF, useInterpolation>;

  using HostGridView = typename Traits::HostGridView;
  using Grid = typename Traits::Grid;
  using GridFunction = GF;
  using IndexSet = typename Traits::IndexSet;
  using Intersection = typename Traits::Intersection;
  using IntersectionIterator = typename Traits::IntersectionIterator;
  using CollectiveCommunication = typename Traits::CollectiveCommunication;

  template <int codim>
  struct Codim
      : public Traits::template Codim<codim>
  {};

  static const bool conforming = Traits::conforming;

public:
  // construct from a host-gridview
  GridView (const Grid& grid, const HostGridView& hostGridView)
    : grid_(&grid)
    , hostGridView_(hostGridView)
  {}

  /// \brief copy constructor. Only copies grid and hostGridView
  GridView (const Self& other)
    : grid_(other.grid_)
    , hostGridView_(other.hostGridView_)
  {}

  /// \brief move constructor. Only moves grid and hostGridView
  GridView (Self&& other)
    : grid_(other.grid_)
    , hostGridView_(std::move(other.hostGridView_))
  {}

  /// \brief copy-assignment operator
  Self& operator= (const Self& other)
  {
    grid_ = other.grid_;
    hostGridView_ = other.hostGridView_;
    if (indexSet_)
      indexSet_.reset(hostGridView().indexSet());
    return *this;
  }

  /// \brief move-assignment operator
  Self& operator= (Self&& other)
  {
    grid_ = other.grid_;
    hostGridView_ = std::move(other.hostGridView_);
    if (indexSet_)
      indexSet_.reset(hostGridView().indexSet()); // TODO: find a better way for copy and move!
    return *this;
  }

  const Grid& grid () const
  {
    assert( grid_ );
    return *grid_;
  }

  const GridFunction& gridFunction () const
  {
    return grid().gridFunction();
  }

  const IndexSet& indexSet () const
  {
    indexSet_.reset(hostGridView().indexSet());
    return indexSet_;
  }

  int size (int codim) const
  {
    return hostGridView().size(codim);
  }

  int size (const GeometryType& type) const
  {
    return hostGridView().size(type);
  }

  template <int codim>
  typename Codim<codim>::Iterator begin () const
  {
    return begin<codim, All_Partition>();
  }

  template <int codim, PartitionIteratorType pit>
  typename Codim<codim>::template Partition<pit>::Iterator begin () const
  {
    using IteratorImp = typename Traits::template Codim<codim>::template Partition<pit>::IteratorImp;
    return IteratorImp::begin(gridFunction(), hostGridView(), grid().order());
  }

  template <int codim>
  typename Codim<codim>::Iterator end () const
  {
    return end<codim, All_Partition>();
  }

  template <int codim, PartitionIteratorType pit>
  typename Codim<codim>::template Partition<pit>::Iterator end () const
  {
    using IteratorImp = typename Traits::template Codim<codim>::template Partition<pit>::IteratorImp;
    return IteratorImp::end(gridFunction(), hostGridView(), grid().order());
  }

  IntersectionIterator ibegin (const typename Codim<0>::Entity& entity) const
  {
    using IteratorImpl = Curved::IntersectionIterator<const Grid, typename HostGridView::IntersectionIterator>;
    return IteratorImpl(entity, hostGridView().ibegin(entity.impl().hostEntity()), grid().order());
  }

  IntersectionIterator iend (const typename Codim<0>::Entity& entity) const
  {
    using IteratorImpl = Curved::IntersectionIterator<const Grid, typename HostGridView::IntersectionIterator>;
    return IteratorImpl(entity, hostGridView().iend(entity.impl().hostEntity()), grid().order());
  }

  const CollectiveCommunication& comm () const
  {
    return hostGridView().comm();
  }

  int overlapSize (int codim) const
  {
    return hostGridView().overlapSize(codim);
  }

  int ghostSize (int codim) const
  {
    return hostGridView().ghostSize(codim);
  }

  template <class DataHandle, class Data>
  void communicate (CommDataHandleIF<DataHandle, Data>& dataHandle,
                    InterfaceType interface,
                    CommunicationDirection direction) const
  {
    using DataHandleIF = CommDataHandleIF<DataHandle, Data>;
    using WrappedDataHandle = Curved::CommDataHandle<Grid, DataHandleIF>;

    WrappedDataHandle wrappedDataHandle(grid(), dataHandle);
    hostGridView().communicate(wrappedDataHandle, interface, direction);
  }

  const HostGridView& hostGridView () const { return hostGridView_; }

private:
  const Grid *grid_;
  HostGridView hostGridView_;
  mutable IndexSet indexSet_;
};

} // end namespace Curved
} // end namespace Dune

#endif // DUNE_CURVEDGRID_GRIDVIEW_HH
